;------------------------------------------------------------------------------
; Name: Custom function macros for MASM
; Version: 1.00
; Author: Petroizki
;------------------------------------------------------------------------------

ifndef PMACROS_INC
	PMACROS EQU 1
	
	__func_flags  = 0
	__func_info   = 0
	__local_bytes = 0
	__local_space = 0
	__stack_bytes = 0
	__timed_func  = 0
	__timer_name TEXTEQU <>
	__timer_var TEXTEQU <>
	
	@ArgRev macro arglist:REQ
		LOCAL txt, arg
		txt TEXTEQU <>
		
		%FOR arg, <arglist>
			txt CATSTR <arg>, <!,>, txt
		ENDM

		txt SUBSTR txt, 1, @SizeStr( %txt ) - 1
		txt CATSTR <!<>, txt, <!>>
		exitm txt
	endm
	
	@IsFloat macro value:REQ
		LOCAL char, isdigit
		
		char SUBSTR <value>, 1, 1
		
		IFIDN char, <!->
			goto find_period
		ELSEIFIDN char, <!0>
			goto find_period
		ELSEIFIDN char, <!1>
			goto find_period
		ELSEIFIDN char, <!2>
			goto find_period
		ELSEIFIDN char, <!3>
			goto find_period
		ELSEIFIDN char, <!4>
			goto find_period
		ELSEIFIDN char, <!5>
			goto find_period
		ELSEIFIDN char, <!6>
			goto find_period
		ELSEIFIDN char, <!7>
			goto find_period
		ELSEIFIDN char, <!8>
			goto find_period
		ELSEIFIDN char, <!9>
			goto find_period
		ELSE
			isdigit = 0
			goto end_macro
		ENDIF
	
		:find_period
		isdigit INSTR <value>, <!.>
		
		:end_macro
		exitm <isdigit>
	endm
	
	;------------------------------------------------------------------------------
	; Name: _pop
	;------------------------------------------------------------------------------
	_pop macro arg:VARARG
		LOCAL argument
		
		%FOR argument, @ArgRev(<arg>)
			pop argument
			
			__stack_bytes = __stack_bytes - 4
		ENDM
	endm
	
	;------------------------------------------------------------------------------
	; Name: _push
	; Example:
	;        _push eax
	;        _push 555.44
	;------------------------------------------------------------------------------
	_push macro arg:VARARG
		LOCAL argument
		
		FOR argument, <arg>
			IF @IsFloat(argument) NE 0
				db 68h
				REAL4 argument
			ELSE
				push argument
			ENDIF
			
			__stack_bytes = __stack_bytes + 4
		ENDM
	endm
	
	_T macro text_string:VARARG
		LOCAL character, const_string, lDeclaration, utext_string
		
		const_string TEXTEQU <text_string>
		
		IF @InStr(1, <text_string>, <!">)
			IFNDEF __amount_macro_strings
				__amount_macro_strings = 0
			ENDIF
			
			lDeclaration CATSTR <STRING_FROM_MACRO>, %__amount_macro_strings
			__amount_macro_strings = __amount_macro_strings + 1
			
			ifndef UNICODE
				;;<code>
				.data
					ALIGN 4
					%lDeclaration db const_string,0
					ALIGN 4
				.code
				;;</code>
			else
				utext_string TEXTEQU <>
				
				FORC character, <const_string>
					IFDIF <character>, <!">
						utext_string CATSTR utext_string, <'>, <character>, <',0,>
					ENDIF
				ENDM
				
				utext_string CATSTR utext_string, <0,0>
				
				;;<code>
				.data
					ALIGN 4
					%lDeclaration db utext_string
					ALIGN 4
				.code
				;;</code>
			endif
		
			exitm <OFFSET lDeclaration>
		ELSE
			exitm <const_string>
		ENDIF
	endm
	
	;------------------------------------------------------------------------------
	; Name: endf
	; Desc: Ends an function.
	; Example:
	;        endf MyFunction
	;------------------------------------------------------------------------------
	endf macro funcname:REQ
		LOCAL func_size, func_sizes, procname
	
		__local_bytes = 0
		__param_bytes = 0
		__stack_bytes = 0
		
		procname TEXTEQU <funcname>
		procname endp
		
		procname CATSTR <__>, procname, <_end>
		procname:
		
		func_size = OFFSET procname - OFFSET funcname
		func_sizes TEXTEQU %func_size
		
		IF func_size GT 8192
			%echo Warning: function funcname greater than 8192 bytes (func_sizes). It may not fit the cache.
		ENDIF
		
		IF __func_info NE 0
			%echo INFO: func funcname, func_sizes bytes.
		ENDIF
		
		OPTION PROLOGUE:PROLOGUEDEF
		OPTION EPILOGUE:EPILOGUEDEF
	endm
	
	;------------------------------------------------------------------------------
	; Name: fncall
	; Desc: Replaces invoke.
	; Example:
	;        fncall function, 5.4, ADDR [LOCAL value], eax
	;        fncall [eax], "My Text"
	;        fncall eax
	;------------------------------------------------------------------------------
	fncall macro function:REQ, parameters:VARARG
		LOCAL bytecount, character, const_string, edxused, paddr, parameter
		
		bytecount = 0
		edxused = 0
		
		IFNB <parameters>
			const_string TEXTEQU <parameters>
		
			%FOR parameter, @ArgRev(<const_string>)
				character SUBSTR <parameter>, 1, 1
				
				IF @IsFloat(parameter) NE 0
					db 68h
					REAL4 parameter
				ELSEIFIDN character, <!">
					push _T(parameter)
				ELSE
					IF @SizeStr(<parameter>) GT 4
						character SUBSTR <parameter>, 1, 5
						
						IFIDNI character, <ADDR >
							paddr SUBSTR <parameter>, 6, @SizeStr(<parameter>) - 5

							IF (OPATTR paddr) AND 00001001b ;; use OFFSET if direct memory address
								push OFFSET paddr
							ELSE
								lea edx, paddr
								push edx
								edxused = 1
							ENDIF
						ELSE
							push parameter
						ENDIF
					ELSE
						IF edxused NE 0
							.ERRIDNI <parameter>, <edx>, <"edx was used with ADDR, and is now fucked up!">
						ENDIF
					
						push parameter
					ENDIF
				ENDIF
			
				bytecount = bytecount + 4
				__stack_bytes = __stack_bytes + 4	
			ENDM
		ENDIF
		
		call function
		__stack_bytes = __stack_bytes - bytecount
	endm
	
	;------------------------------------------------------------------------------
	; Name: fnjmp
	; Desc: The same as fncall, except return address can be defined custom.
	;        If return address is 0, no value is pushed to the stack, so return
	;        will crash.
	; Example:
	;        fnjmp <mylabel>, func0, 55 ; calls func0, but returns to 'mylabel'
	;        fnjmp <0>, func0, 55 ; calls func0, but no return address is defined
	;------------------------------------------------------------------------------
	fnjmp macro retaddr:REQ, function:REQ, parameters:VARARG
		LOCAL bytecount, character, const_string, edxused, paddr, parameter
		
		bytecount = 0
		edxused = 0
		
		IFNB <parameters>
			const_string TEXTEQU <parameters>
		
			%FOR parameter, @ArgRev(<const_string>)
				character SUBSTR <parameter>, 1, 1
				
				IF @IsFloat(parameter) NE 0
					db 68h
					REAL4 parameter
				ELSEIFIDN character, <!">
					push _T(parameter)
				ELSE
					IF @SizeStr(<parameter>) GT 4
						character SUBSTR <parameter>, 1, 5
						
						IFIDNI character, <ADDR >
							paddr SUBSTR <parameter>, 6, @SizeStr(<parameter>) - 5

							IF (OPATTR paddr) AND 00001001b ;; use OFFSET if direct memory address
								push OFFSET paddr
							ELSE
								lea edx, paddr
								push edx
								edxused = 1
							ENDIF
						ELSE
							push parameter
						ENDIF
					ELSE
						IF edxused NE 0
							.ERRIDNI <parameter>, <edx>, <"edx was used with ADDR, and is now fucked up!">
						ENDIF
					
						push parameter
					ENDIF
				ENDIF
			
				bytecount = bytecount + 4
				__stack_bytes = __stack_bytes + 4	
			ENDM
		ENDIF
		
		IF (OPATTR retaddr) AND 11011011b		; if is mem address etc..
			push retaddr
		ELSEIF (OPATTR retaddr) AND 00000100b		; if is constant
			IF retaddr EQ 0
								; do nothing..
			ELSE
				push retaddr
			ENDIF
		ELSE
			push retaddr
		ENDIF
		
		jmp function
		__stack_bytes = __stack_bytes - bytecount
	endm
	
	;------------------------------------------------------------------------------
	; Name: func
	; Desc: Function definition, replaces the proc. Using this also makes the
	;        invoke useless, you must the new macro fncall to call functions.
	;        Use the macro endf to end a function.
	; Example:
	;        func function, MOVE, stdcall uses ebx, dwValue:DWORD, lpRect:ADDR RECT
	;------------------------------------------------------------------------------
	func macro funcname:REQ, args:VARARG
		LOCAL arg, parambytes, paramname, paramtype, procdef, procname, pos, temp0
		
		OPTION EPILOGUE:FUNCEPILOGUE
		OPTION PROLOGUE:FUNCPROLOGUE
		OPTION NOKEYWORD:<invoke LOCAL>
		LOCAL TEXTEQU <esp + __stack_bytes - >
		PARAM TEXTEQU <esp + __stack_bytes + >
		
		__func_align  = 1
		__func_flags  = 0
		__func_frames = 0
		__func_info   = 0
		__move_frames = 0
		__stack_bytes = 0
		__time_func   = 0
		
		parambytes = 4
		;; this string will be passed to original proc macro
		procdef TEXTEQU <>
		
		FOR arg, <args>
			pos INSTR <arg>, <!:>
			
			IF pos GT 1		
				paramname SUBSTR <arg>, 1, pos - 1	;; get parameter name
				paramtype SUBSTR <arg>, pos + 1, @SizeStr(arg) - pos	;; get param type
				%paramname  = parambytes
				
				IF @SizeStr(%paramtype) GT 4
					temp0 SUBSTR paramtype, 1, 5	;; get first five chars
					
					IFIDNI temp0, <ADDR >			;; check if this is address
						parambytes = parambytes + 4
						paramtype TEXTEQU <DWORD>	;; this is for procdef
					ELSE
						parambytes = parambytes + TYPE paramtype
					ENDIF
				ELSE
					parambytes = parambytes + TYPE paramtype
				ENDIF
				
				;; add the stuff for proc definition
				procdef CATSTR procdef, <!,>, <__ebp__>, paramname, <!:>, <paramtype>
			ELSE
				IFIDNI <arg>, <FRAMES>
					__func_frames = 1
					LOCAL TEXTEQU <ebp - >
					PARAM TEXTEQU <ebp + 4 + >
				ELSEIFIDNI <arg>, <INFO>
					__func_info = 1
				ELSEIFIDNI <arg>, <TIME>
					__time_func = 1
				ELSEIFIDNI <arg>, <MOVE>
					__move_frames = 1
				ELSEIFIDNI <arg>, <NOALIGN>
					__func_align = 0
				ELSE
					;; just add the text to proc string
					procdef CATSTR procdef, <! >, <arg>
				ENDIF
			ENDIF
		endm
		
		IF parambytes EQ 4
			IF __local_bytes EQ 0
				__func_frames = 0
			ENDIF
		ENDIF

		IFE __func_align EQ 0
			ALIGN 4
		ENDIF

		;; can't use funcname directly here, it will hit error on proc names starting with 'out'
		procname TEXTEQU <funcname>
		procname proc procdef
	endm
	
	;------------------------------------------------------------------------------
	;------------------------------------------------------------------------------
	FUNCEPILOGUE macro procname, flag, parmbytes, localbytes, reglist, userparms
		LOCAL reg, regbytes
		
		regbytes = 0
		
		IF __func_frames EQ 0
			IF __move_frames EQ 0
				;; preserve used registers
				FOR reg, reglist
					pop reg
				ENDM
			
				IFE __local_bytes EQ 0
					IF __local_bytes LE 4
						pop ecx
					ELSE
						add esp, __local_bytes
					ENDIF
				ENDIF
			ELSE
				regbytes = 0
				
				FOR reg, reglist
					mov reg, dword ptr [esp + regbytes]
					regbytes = regbytes + 4
				ENDM
				
				regbytes = regbytes + __local_bytes
				
				IFE regbytes EQ 0
					IF regbytes LE 4
						pop ecx
					ELSE
						add esp, regbytes
					ENDIF
				ENDIF
			ENDIF
		ELSE
			IF __move_frames EQ 0
				;; preserve used registers
				FOR reg, reglist
					pop reg
				ENDM
				
				IFE __local_bytes EQ 0
					mov esp, ebp
				ENDIF
			ELSE
				FOR reg, reglist
					regbytes = regbytes + 4
				ENDM
				
				regbytes = regbytes + __local_bytes
				
				IFE regbytes EQ 0
					add esp, regbytes
				ENDIF
				
				FOR reg, reglist
					mov reg, dword ptr [ebp - regbytes]
					regbytes = regbytes - 4
				ENDM
			ENDIF
			
			pop ebp
		ENDIF

		;; epilogue code for TIMED FUNCTION
		IFE __time_func EQ 0
			;; <code>
;			pushad
;			rdtsc
;			sub eax, __func_time0
;			fncall DwordToSz10, eax, ADDR __timer_var
;			fncall OutputDebugString, ADDR __timer_name
;			popad
			;; </code>
		ENDIF
		
		;; correct stack frames or not?
		IF (flag AND 010000y) OR (parmbytes EQ 0)	;; check if C calling convention is set
			ret										;; return, caller will clear the stack
		ELSE
			ret parmbytes							;; return and clear stack
		ENDIF
	endm
	
	;------------------------------------------------------------------------------
	;------------------------------------------------------------------------------
	FUNCPROLOGUE macro procname, flag, parmbytes, localbytes, reglist, userparms
		LOCAL reg, regbytes, TIMER_NAME

		regbytes = 0
		
		;; prologue code for TIMED FUNCTION
		IFE __time_func EQ 0
			;; the following code is executed only once, when the first, timed function is encountered
			IF __timed_func EQ 0
				;; <code>
;				.data?
;				ALIGN 4
;				__func_time0 dd ?
;				.code
				;; </code>
				
				__timed_func = 1
			ENDIF
			
			;; <code>
;			__timer_name CATSTR <__>, <procname>, <_name>
;			__timer_var CATSTR <__>, <procname>, <_var>
;			TIMER_NAME CATSTR <!">, <procname>, <: ">
;			.data
;			__timer_name db TIMER_NAME
;			__timer_var db 16 dup (?)
;			.code
			
;			pushad
;			rdtsc
;			mov __func_time0, eax
;			popad
			;; </code>
		ENDIF
		
		IF __func_frames EQ 0
			IF __move_frames EQ 0
				IFE __local_bytes EQ 0
					IF __local_bytes LE 4
						push ecx
					ELSE
						sub esp, __local_bytes
					ENDIF
					
					__stack_bytes = __stack_bytes + __local_bytes
					__local_space = __stack_bytes
				ENDIF
			
				;; preserve used registers
				FOR reg, reglist
					.ERRE TYPE reg EQ 4, <"ERROR: All preserved registers should be 4bytes.">
					
					push reg
					__stack_bytes = __stack_bytes + 4
				ENDM
			ELSE
				regbytes = __local_bytes
				
				FOR reg, reglist
					regbytes = regbytes + 4
					mov dword ptr [esp - regbytes], reg
				ENDM
				
				__stack_bytes = __stack_bytes + regbytes - __local_bytes
				
				IFE regbytes EQ 0
					IF regbytes LE 4 AND __local_bytes GE 4
						push ecx
					ELSE
						sub esp, regbytes
					ENDIF
					
					__stack_bytes = __stack_bytes + __local_bytes
					__local_space =  __stack_bytes
				ENDIF
			ENDIF
		ELSE
			push ebp
			mov ebp, esp
			
			IF __move_frames EQ 0
				IFE __local_bytes EQ 0
					IF __local_bytes LE 4
						push ecx
					ELSE
						sub esp, __local_bytes
					ENDIF
				ENDIF
			
				;; preserve used registers
				FOR reg, reglist
					.ERRE TYPE reg EQ 4, <"ERROR: All preserved registers should be 4bytes.">
					.ERRIDNI <reg>, <ebp>, <"ERROR: Can't preserve ebp on frames.">
					push reg
				ENDM
			ELSE
				FOR reg, reglist
					regbytes = regbytes + 4
				ENDM
				
				regbytes = regbytes + __local_bytes
				
				IFE regbytes EQ 0
					sub esp, regbytes
				ENDIF
				
				regbytes = -__local_bytes
			
				FOR reg, reglist
					.ERRIDNI <reg>, <ebp>, <"ERROR: Can't preserve ebp on frames.">
					regbytes = regbytes - 4
					mov dword ptr [ebp + regbytes], reg
				ENDM
			ENDIF
		ENDIF
		
		exitm %parmbytes
	endm
	
	;------------------------------------------------------------------------------
	; Name: LOCALS
	; Desc: Defines local variables for function. Locals are defined before the
	;        function. To refer local in a function, use the keyword 'LOCAL'.
	; Example:
	;        LOCALS dwTemp:DWORD, array[4]:BYTE
	;        func function
	;         mov dword ptr [LOCAL dwTemp], eax ; moves eax to local dwTemp
	;------------------------------------------------------------------------------
	LOCALS macro args:VARARG
		LOCAL amount, arg, localequ, localname, localtype, pos0, pos1, temp
		__local_bytes = 0
		
		IFB <args>
			exitm			;; quit macro if no args defined
		ENDIF
		
		%FOR arg, <args>		;; go trough each local
			amount TEXTEQU <!1>
		
			pos0 INSTR <arg>, <!:>	;; find ':'
			
			.ERRE pos0 GT 1, <"Error: LOCALS are messsed up, make sure the name and type is defined.">
			
			localname SUBSTR <arg>, 1, pos0 - 1	;; get local name
			localtype SUBSTR <arg>, pos0 + 1, @SizeStr(<arg>) - pos0 ;; get local type
			
			pos0 INSTR localname, <![>
			
			IF pos0 GT 1
				pos1 INSTR localname, <!]>
				.ERRE pos1 EQ @SizeStr(%localname), <"Error: LOCALS are messsed up, make sure the name and type is defined.">
				.ERRE pos1 GT pos0, <"Error: LOCALS are messsed up, make sure the name and type is defined.">
				
				amount SUBSTR localname, pos0 + 1, pos1 - pos0 - 1
				localname SUBSTR localname, 1, pos0 - 1
			ENDIF
			
			.ERRE amount GT 0, <"Error: Size of local cannot be 0.">
			
			temp = (TYPE localtype) * amount
			__local_bytes = __local_bytes + temp
			
			temp = temp AND 3
			
			IF temp NE 0
				temp = 4 - temp
			ENDIF
			
			__local_bytes = __local_bytes + temp
			%localname = __local_bytes
		ENDM
	endm
	
	;------------------------------------------------------------------------------
	; Name: popad
	;------------------------------------------------------------------------------
	_popad macro
		db 61h ;; popad
		
		__stack_bytes = __stack_bytes - 32
	endm
	
	;------------------------------------------------------------------------------
	; Name: pushad
	;------------------------------------------------------------------------------
	_pushad macro
		db 60h ;; pushad
		
		__stack_bytes = __stack_bytes + 32
	endm
	
	;------------------------------------------------------------------------------
	; Name: popfd
	;------------------------------------------------------------------------------
	_popfd macro
		db 9Dh ;; popfd
		
		__stack_bytes = __stack_bytes - 4
	endm
	
	;------------------------------------------------------------------------------
	; Name: pushfd
	;------------------------------------------------------------------------------
	_pushfd macro
		db 9Ch ;; pushfd
		
		__stack_bytes = __stack_bytes + 4
	endm
endif